import pygame
from classes.config import Config
from classes.event_stack import EventStack, Event, WindowStack


class Box:
    """

    Attributes:
        size (tuple): A tuple containing the width and height of the box.
        close_button (bool): Indicates if the box should have a close button. Default is True.
        close_callback (callable): The callback function to be executed when the box is closed.
        parent (Box): The parent class, if this box is a child. Default is None.
        Surface (pygame.Surface): The surface on which the box is drawn.
        position (tuple): The position of the box on the screen.
        opened (bool): Indicates if the box is currently open.
    """

    def __init__(self, size, close_button=True, close_callback=None, parent=None) -> None:
        """Initializes a new Box instance with the specified size and options."""
        self.size = size
        self._close_callback = None
        self.close_button = close_button
        self.close_callback = close_callback
        self.parent = parent
        self.Surface = pygame.Surface(size, pygame.SRCALPHA)
        self.Surface.fill((254, 211, 127))
        self.add_border()

        self.position = (0, 0)
        self._opened = False

        self.event_mouse_on = Event("Box_close_mouse_on", self.mouse_on_close)
        self.event_close = Event("close", self.close)

    @property
    def opened(self):
        return self._opened

    @opened.setter
    def opened(self, value):
        if value:
            self._opened = value
            if self.close_button:
                EventStack.push(self.event_mouse_on)
            EventStack.push(self.event_close)
            WindowStack.add_window(self)

            if self.parent:
                self.parent.opened = True
        else:
            self.close()

    def reset_surface(self):
        """Resets the surface of the box and fills it with its default color and border."""
        self.Surface = pygame.Surface(self.size, pygame.SRCALPHA)
        self.Surface.fill((254, 211, 127))
        self.add_border()

    @property
    def close_callback(self):
        """The callback function to be executed when the box is closed."""
        return self._close_callback

    @close_callback.setter
    def close_callback(self, value):
        """Sets the callback function to be executed when the box is closed."""
        self._close_callback = value

    def add_element(self, element, pos):
        """Adds a graphical element to the box at the specified position."""
        self.Surface.blit(element, pos)

    def reset_and_add(self, element, pos):
        """Resets the surface of the box and adds a graphical element at the specified position."""
        self.reset_surface()
        self.Surface.blit(element, pos)

    def add_border(self):
        """Adds a border to the box."""
        pygame.draw.rect(self.Surface, (51, 32, 24),
                         (0, 0, self.size[0], self.size[1]), 2, 2, 2, 2)
        pygame.draw.rect(self.Surface, (76, 48, 36),
                         (0, 0, self.size[0], 2))

        pygame.draw.rect(self.Surface, (153, 69, 63),
                         (2, 2, self.size[0]-4, self.size[1]-4), 4, 4, 4, 4)

        pygame.draw.rect(self.Surface, (51, 32, 24),
                         (4, 4, self.size[0]-8, self.size[1]-8), 2, 2, 2, 2)

        if self.close_button:
            pygame.draw.rect(self.Surface, (254, 211, 127),
                             (self.size[0]-20, 0, 20, 20))
            pygame.draw.line(self.Surface, (0, 0, 0),
                             (self.size[0]-20, 0), (self.size[0], 20), 2)
            pygame.draw.line(self.Surface, (0, 0, 0),
                             (self.size[0], 0), (self.size[0]-20, 20), 2)

    def set_close_event(self, event):
        """Sets the close event callback function."""
        self.close_callback = event

    def close(self):
        """Closes the box and executes the close callback function if present."""
        self._opened = False
        if self._close_callback:
            self._close_callback()

        if self.parent:
            self.parent.opened = False
        EventStack.remove(self.event_mouse_on)
        EventStack.remove(self.event_close)

    def mouse_on_close(self):
        """Handles mouse events when the mouse hovers over the close button."""
        mouse_pos = pygame.mouse.get_pos()
        if (mouse_pos[0] > self.position[0] + self.size[0]-20 and mouse_pos[0] < self.position[0] + self.size[0]
                and mouse_pos[1] > self.position[1] and mouse_pos[1] < self.position[1] + 20):
            Config.cursor_style = pygame.SYSTEM_CURSOR_HAND
            if pygame.mouse.get_pressed()[0]:

                self.close()

    def __call__(self, position=None):
        """Returns the surface of the box and updates its position if specified."""
        if self.close_button:
            if not EventStack.find_event(self.event_mouse_on):
                EventStack.push(self.event_mouse_on)
        if not EventStack.find_event(self.event_close):
            EventStack.push(self.event_close)

        if position:
            self.position = position

        return self.Surface



from classes.item import Item, MAPCOLOR
from typing import MutableSequence
import pygame
import random
from classes.map_reader import get_map_sprite_image
from classes.config import Config


class Block:
    """Block class"""

    size = 40

    def __init__(self, coords, block_type: MAPCOLOR) -> None:
        """ Block constructor """

        self.items: MutableSequence[Item] = []
        self.coords = coords
        self.indexes = self.calc_indexes()
        self.type = block_type
        self.image = pygame.Surface((Block.size, Block.size))
        self.display_text = ""
        self.reset_block_image()

    def calc_indexes(self):
        """Calculates the indexes of the block in the map """
        return (self.coords[0] // Block.size, self.coords[1] // Block.size)

    def reset_block_image(self):
        """Resets the block image to the default"""
        self.image.fill(MAPCOLOR.GRASS.rgb(MAPCOLOR.GRASS.value))
        self.image.blit(get_map_sprite_image(
            pygame.image.load('assets/images/grass.png').convert_alpha(), (random.randint(0, 4), random.randint(0, 4))), (4, 4))

    def add_item(self, item: Item) -> None:
        """Adds an item to the block"""
        self.items.append(item)

        item_type = ""
        match item.type:
            case "BERRY" | "MUSHROOM" | "APPLE" | "CARROT":
                item_type = "FOOD"
            case _:
                item_type = item.type

        Config.items[item_type].append(self)

        self.set_item_image()

    def remove_item_from_top(self):
        """ Removes the top item from the block """
        if len(self.items) > 0:
            removed_item = self.items.pop()
            self.set_item_image()

            item_type = ""
            match removed_item.type:
                case "BERRY" | "MUSHROOM" | "APPLE" | "CARROT":
                    item_type = "FOOD"
                case _:
                    item_type = removed_item.type

            Config.remove_from_items(item_type, self)
            return removed_item
        return None

    def set_item_image(self):
        """Sets the block image to the top item image"""
        self.reset_block_image()
        for item in self.items:
            self.image.blit(item.item_image, (4, 4))

    def draw(self, screen):
        """Draws the block on the screen"""
        self.set_item_image()
        screen.blit(self.image, self.coords)

    def on_block_check(self, coords):
        """Checks if the mouse is on the block"""
        if self.coords[0] <= coords[0] <= self.coords[0] + Block.size and self.coords[1] <= coords[1] <= self.coords[1] + Block.size:
            return self

        return None


class GameMap:

    def __init__(self) -> None:
        self.blocks: MutableSequence[Block] = []

    def add_block(self, block: Block) -> None:
        self.blocks.append(block)

    def draw(self, screen):
        for block in self.blocks:
            block.draw(screen)

    def on_block_check(self, coords) -> Block:
        index_x = coords[0] // Block.size
        index_y = coords[1] // Block.size
        block = self.blocks[int(index_x*100+index_y)]

        if block:
            item = block.on_block_check(coords)
            if item is not None:
                return item
        return None

    def get_block_by_coords(self, coords) -> Block:
        for block in self.blocks:
            if block.coords[0] <= coords[0] <= block.coords[0] + Block.size and block.coords[1] <= coords[1] <= block.coords[1] + Block.size:
                return block
        return None

    def get_block_by_indexes(self, indexes) -> Block:
        for block in self.blocks:
            if block.indexes == indexes:
                return block
        return None





import pygame
from classes.config import Config


def get_item_sprite_image(sprite_sheet: pygame.Surface, row_index: int, count: int, sprite_size=32, spacing: int = 0):
    """Retruns a list of sprites from a sprite sheet"""
    sprites = [pygame.Surface(
        (sprite_size, sprite_size), pygame.SRCALPHA) for x in range(count)]

    for i in range(count):

        sprites[i].blit(sprite_sheet, (0, 0), (i * (sprite_size + spacing), row_index * (sprite_size + spacing),
                                               sprite_size, sprite_size))

    return sprites


def scale_sprites(sprites: list, scale: float):
    """ Returns a list of scaled sprites """
    return [pygame.transform.scale(x, (x.get_width() * scale, x.get_height() * (scale*Config.data['scale_multiplier']))) for x in sprites]


def scale_image(image: pygame.Surface, scale: float):
    """ Returns a scaled image """
    return pygame.transform.scale(image, (image.get_width() * scale, image.get_height() * (scale*Config.data['scale_multiplier'])))




from ordered_set import OrderedSet


class Event:
    def __init__(self, event_type: str, func, *args) -> None:
        self.type = event_type
        self.func = func
        self.args = args

    def __call__(self):
        self.func(*self.args)

    def __str__(self) -> str:
        return self.type + " " + str(self.func)


class EventStack:

    stack = OrderedSet()

    @staticmethod
    def push(event):

        EventStack.stack.add(event)

    @staticmethod
    def pop(index=-1):

        return EventStack.stack.pop(index)

    @staticmethod
    def remove(event):
        try:
            EventStack.stack.discard(event)
        except KeyError:
            print("Not in")

    @staticmethod
    def peek(index=-1):

        return EventStack.stack[index]

    @staticmethod
    def find(event_type):

        for event in EventStack.stack:

            if event.type == event_type:

                return event

        return None

    @staticmethod
    def find_event(event):

        for e in EventStack.stack:

            if e == event:

                return e

        return None

    @staticmethod
    def find_and_call(event_type):

        event = next(
            (x for x in EventStack.stack[::-1] if x.type == event_type), None)

        if event is not None:

            event()

    @staticmethod
    def iterate():

        for event in EventStack.stack:

            event()

    @staticmethod
    def call_last():

        event = EventStack.stack[-1]
        event()

    @staticmethod
    def call_and_pop(event_type):
        event = next(
            (x for x in EventStack.stack[::-1] if x.type == event_type), None)

        if event is not None:

            EventStack.remove(event)
            event()


class WindowStack:

    stack = []

    @staticmethod
    def add_window(window):
        WindowStack.stack.append(window)

    @staticmethod
    def remove_window(window):
        WindowStack.stack.remove(window)





import pygame


class TextDisplay:
    def __init__(self, text, font_size, color, outline_color=(0, 0, 0)) -> None:
        self.text = text
        self.font_size = font_size
        self.color = color
        self.outline_color = outline_color

        self.font = pygame.font.SysFont("Arial", self.font_size, bold=True)
        self.text_surface = self.font.render(
            self.text, True, self.color)
        self.text_rect = self.text_surface.get_rect()

        self.outline_surface = self.font.render(
            self.text, True, self.outline_color)

        self.Surface = self.draw()

    def draw(self):
        surface = pygame.Surface(
            (self.text_rect.width+1, self.text_rect.height+1), pygame.SRCALPHA)
        surface.blit(self.outline_surface, (1, 1))
        surface.blit(self.text_surface, (0, 0))
        return surface


class PlayerInfoText:
    def __init__(self, coords) -> None:
        self.stack = []
        self.default_coords = coords

    def add(self, text: TextDisplay, duration: int):
        self.stack.append(
            {"text": text, "duration": duration})

    def draw(self, screen: pygame.Surface):
        for text in self.stack:
            screen.blit(text["text"].draw(), self.default_coords)
            text["duration"] -= 1
            self.default_coords = (
                self.default_coords[0], self.default_coords[1] - 20)
            if text["duration"] <= 0:
                self.stack.remove(text)
                self.default_coords = (
                    self.default_coords[0], self.default_coords[1] + 20)

    def set_coords(self, coords: tuple):
        self.default_coords = coords



from setuptools import setup

setup(
    name="CatRoyale",
    version="1.5",
    description="A cat themed battle royale game",
    author="Tarsoly Barnabás",
    author_email="tarsoly.barnabas2002@gmail.com",
    url="https://github.com/Iseroo/Python-kotprog",
    packages=["classes", "assets"],
    py_modules=["main", "__main__"],
    install_requires=["pygame", "webcolors", "PIL", "ordered_set"],
    include_package_data=True,
)





class MessageService:
    messages = []

    @staticmethod
    def add(message: dict) -> None:
        MessageService.messages.append(message)

    @staticmethod
    def get_messages():
        return MessageService.messages

    @staticmethod
    def next() -> dict:
        return MessageService.messages.pop(0) if len(MessageService.messages) > 0 else None
